/*
 * animatch.c: animate images together.
 *
 * Copyright 2002 by Akkana Peck.
 * You are free to use or modify this code under the Gnu Public License.
 */

#include <stdlib.h>       // for getenv()
#include <stdio.h>

#include <math.h>

#include <X11/Xlib.h>
#include <X11/xpm.h>
#include <X11/keysym.h>

static Display *dpy;
static int screen;
static Window win;
static GC gc;

static int XWinSize = 500;
static int YWinSize = 300;
static XImage* image;

static int ArgC, ArgP;
static char** ArgV;

static long EventMask = ( ExposureMask
                          | ButtonPressMask| ButtonReleaseMask
                          | KeyPressMask
                          | StructureNotifyMask );

typedef struct XY_s
{
    int x, y;
} XY;
#define NPOINTS 20
static XY Points[NPOINTS];
static XY FirstPoints[NPOINTS];
static int NumPoints = 0;
static int NumFirstPoints = 0;
static XY FirstSize;

void PrintPoints()
{
    int i;
    printf("%s (%d x %d):", ArgV[ArgP], XWinSize, YWinSize);
    for (i=0; i<NumPoints; ++i)
        printf(" (%d, %d)", Points[i].x, Points[i].y);
    printf("\n");

    if (ArgP > 1 && NumPoints >= 2 && NumFirstPoints >= 2)
    {
        // Calculate how to transform this image to match the first.
        // For now, we only take into account the first two points.
        // First, calculate rotation.
        int dy0 = FirstPoints[1].y - FirstPoints[0].y;
        int dy1 = Points[1].y - Points[0].y;
        int dx0 = FirstPoints[1].x - FirstPoints[0].x;
        int dx1 = Points[1].x - Points[0].x;
        double angle = atan2(dy1, dx1) - atan2(dy0, dx0);
        double r, theta;
        double xx, yy, xt, yt;
        if (angle > M_PI) angle -= M_PI;

        printf("\tRotate %s by %d degrees\n",
               angle > 0 ? "ccw" : "cw",
               abs((int)(angle * 180. / M_PI)));

        // If we rotated, then the image size has increased,
        // so figure out by how much:
        xt = XWinSize * sin(angle);
        yt = YWinSize * sin(angle);  // oddly enough, they're both sin

        // Now, figure out how to translate the coordinates
        // after a rotation around the center.
        // First, translate the two current points to radial:
        xx = Points[0].x - XWinSize/2;
        yy = Points[0].y - YWinSize/2;
        r = sqrt(xx*xx + yy*yy);
        theta = atan2(yy, xx) + angle;   // theta after rotation
        // Now translate back:
        xx = r * cos(theta) + XWinSize/2 + xt;
        yy = r * sin(theta) + YWinSize/2 + yt;
        // xx and yy are relative to the center.
        // They are the coordinates of the zeroth point
        // in the coordinate system of the rotated and hence bigger image.
        printf("\tCut off the first (%d, %d)\n",
               (int)xx - FirstPoints[0].x, (int)xx - FirstPoints[0].y);
    }

    NumPoints = 0;
}

void ShowImage()
{
    int status;
    XImage* shapeImage;
    XpmAttributes attributes;

    attributes.valuemask = 0;
    status = XpmReadFileToImage(dpy, ArgV[ArgP],
                                &image, &shapeImage, &attributes);
    if (status)
    {
        fprintf(stderr, "Error reading file %s\n", ArgV[ArgP]);
        return;
    }
    XWinSize = image->width;
    YWinSize = image->height;
    if (ArgP == 1) {
        FirstSize.x = XWinSize;
        FirstSize.y = YWinSize;
    }

    // Turn off event processing, else we get an infinite Expose loop:
    XResizeWindow(dpy, win, XWinSize, YWinSize);
    XPutImage(dpy, win, gc, image, 0, 0, 0, 0, XWinSize, YWinSize);
    XSync(dpy, 1);
}

int HandleEvent()
{
    XEvent event;
    char buffer[20];
    KeySym keysym;
    XComposeStatus compose;

    XNextEvent(dpy, &event);
    switch (event.type)
    {
      case Expose:
          //printf("Expose\n");
          ShowImage();
          break;
      case ConfigureNotify:
          XWinSize = event.xconfigure.width;
          YWinSize = event.xconfigure.height;
          //printf("ConfigureNotify: now (%d, %d)\n", XWinSize, YWinSize);
          //XSelectInput(dpy, win, EventMask);
          break;
      case ButtonPress:
          Points[NumPoints].x = event.xbutton.x;
          Points[NumPoints].y = event.xbutton.y;
          if (ArgP == 1)
          {
              FirstPoints[NumFirstPoints].x = Points[NumPoints].x;
              FirstPoints[NumFirstPoints].y = Points[NumPoints].y;
              ++NumFirstPoints;
          }
          ++NumPoints;
          break;
      case ButtonRelease:
          break;
      case KeyPress:
          XLookupString(&(event.xkey), buffer, sizeof buffer,
                        &keysym, &compose);
          switch (keysym)
          {
            case XK_space:
                // Go to next image
                PrintPoints();
                if (ArgP >= ArgC-1)
                    return -1;
                ++ArgP;
                ShowImage();
                break;
            case XK_BackSpace:
                if (ArgP > 1)
                {
                    PrintPoints();
                    --ArgP;
                    ShowImage();
                }
                break;
            case XK_q:
                PrintPoints();
                return -1;
            default:
                break;
          }
          break;

      case MapNotify:
      case ReparentNotify:
          break;
      default:
          printf("Unknown event: %d\n", event.type);
    }
    return 0;
}

int main(int argc, char** argv)
{
    if (argc <= 1) {
        printf("Usage: blah blah\n");
        return 1;
    }
    ArgV = argv;
    ArgC = argc;
    ArgP = 1;

    if ((dpy = XOpenDisplay(getenv("DISPLAY"))) == 0)
    {
        fprintf(stderr, "Can't open display: %s\n", getenv("DISPLAY"));
        exit(1);
    }
    screen = DefaultScreen(dpy);

    win = XCreateSimpleWindow(dpy, RootWindow(dpy, screen),
                              0, 0, XWinSize, YWinSize, 3,
                              WhitePixel(dpy, screen),
                              BlackPixel(dpy, screen));
    if (!win)
    {
        fprintf(stderr, "Can't create window\n");
        exit(1);
    }

    gc = XCreateGC(dpy, win,  0, 0);
    if (!gc) {
        fprintf(stderr, "Couldn't create gc!\n");
        return 1;
    }

    XSelectInput(dpy, win, EventMask);

    ShowImage();
    XMapWindow(dpy, win);

    while (HandleEvent() >= 0)
        ;
    return 0;
}

